import boto3
import json
import os

from boto3utils import s3
from boto3.dynamodb.conditions import Key
from datetime import datetime, timedelta
from logging import getLogger
from traceback import format_exc
from typing import Dict, Optional, List

logger = getLogger(__name__)
logger.setLevel(os.getenv('CIRRUS_LOG_LEVEL', 'INFO'))

# envvars
CATALOG_BUCKET = os.getenv('CIRRUS_CATALOG_BUCKET')

STATES = ['QUEUED', 'PROCESSING', 'COMPLETED', 'FAILED', 'INVALID']
INDEX_KEYS = {
    'input_state': 'input_collections',
    'output_state': 'output_collections'
}


class StateDB:

    def __init__(self, table_name: str=os.getenv('CIRRUS_STATE_DB', 'test')):
        """Initialize a StateDB instance using the Cirrus State DB table

        Args:
            table_name (str, optional): The Cirrus StateDB Table name. Defaults to os.getenv('CIRRUS_STATE_DB', None).
        """
        # initialize client
        self.db = boto3.resource('dynamodb')
        self.table_name = table_name
        self.table = self.db.Table(table_name)

    def create_item(self, catalog: Dict, state: str='QUEUED'):
        """Create an item in DynamoDB

        Args:
            catalog (Dict): A Cirrus Input Catalog
            state (str, optional): Set items to this state. Defaults to 'QUEUED'.
        """
        now = datetime.now().isoformat()
        opts = catalog['process']['output_options']
        output_collections = '/'.join(sorted(opts['collections'].keys()))
        key = self.catid_to_key(catalog['id'])
        response = self.table.put_item(
            Item={
                'input_collections': key['input_collections'],
                'id': key['id'],
                'output_collections': output_collections,
                'current_state': f"{state}_{now}",
                'created_at': now,
            }
        )
        logger.debug(f"Created DynamoDB Item {catalog['id']}")
        return response

    def delete_item(self, catid: str):
        key = self.catid_to_key(catid)
        response = self.table.delete_item(Key=key)
        logger.debug(f"Removed DynamoDB Item {catid}")
        return response

    def get_dbitem(self, catid: str) -> Dict:
        """Get a DynamoDB item

        Args:
            catid (str): Catalog ID

        Raises:
            Exception: Error getting item

        Returns:
            Dict: DynamoDB Item
        """
        try:
            response = self.table.get_item(Key=self.catid_to_key(catid))
            logger.debug(f"Fetched {response['Item']}")
            return response['Item']
        except Exception as err:
            logger.info(f"Error fetching item {catid}: {err}")
            # no such item
            return None

    def get_dbitems(self, catids: List[str]) -> List[Dict]:
        """Get multiple DynamoDB Items

        Args:
            catids (List[str]): A List of Catalog IDs

        Raises:
            Exception: Error getting items

        Returns:
            List[Dict]: A list of DynamoDB Items
        """
        try:
            resp = self.db.meta.client.batch_get_item(RequestItems={
                self.table_name: {
                    'Keys': [self.catid_to_key(id) for id in catids]
                }
            })
            items = []
            for r in resp['Responses'][self.table_name]:
                items.append(r)
            logger.debug(f"Fetched {len(items)} items")
            return items
        except Exception as err:
            msg = f"Error fetching items {catids} ({err})"
            logger.error(msg)
            logger.error(format_exc())
            raise Exception(msg) from err

    def get_counts(self, collection: str, state: str=None, since: str=None,
                   index: str='input_state', limit: int=None) -> Dict:
        """Get counts by query

        Args:
            collection (str): /-separated list of collections (input or output depending on index)
            state (Optional[str], optional): State of Items to get. Defaults to None.
            since (Optional[str], optional): Get Items since this amount of time in the past. Defaults to None.
            index (str, optional): Query this index (input_state or output_state). Defaults to 'input_state'.
            limit (int, optional): The max number to return, anything over will be reported as "<limit>+", e.g. "1000+"

        Returns:
            Dict: JSON containing counts key with counts for each state requested
        """
        counts = {}

        # make sure valid collection
        assert(index in INDEX_KEYS.keys())

        _states = [state] if state else STATES

        for state in _states:
            counts[state] = 0
            resp = self.query(collection, state, since=since, index=index, select='COUNT')
            counts[state] = resp['Count']
            while 'LastEvaluatedKey' in resp:
                resp = self.query(collection, state, since=since, index=index, select='COUNT',
                                         ExclusiveStartKey=resp['LastEvaluatedKey'])
                counts[state] += resp['Count']
                if limit and counts[state] > limit:
                    break
            if limit and counts[state] > limit:
                counts[state] = f"{limit}+"
                continue
        return {
            INDEX_KEYS[index]: collection,
            'index': index,
            'counts': counts
        }

    def get_items_page(self, collection: str, state: str, since: Optional[str]=None,
                  index: str='input_state', limit=100, nextkey: str=None) -> List[Dict]:
        """Get Items by query

        Args:
            collection (str): /-separated list of collections (input or output depending on index)
            state (str): State of Items to get (QUEUED, PROCESSING, COMPLETED, FAILED, INVALID)
            since (Optional[str], optional): Get Items since this amount of time in the past. Defaults to None.
            index (str, optional): Query this index (input_state or output_state). Defaults to 'input_state'.

        Returns:
            Dict: List of Items
        """
        if state:
            _states = [state]
        _states = [state] if state else STATES

        items = {
            'items': []
        }
        if nextkey:
            dbitem = self.get_dbitem(nextkey)
            startkey = { key: dbitem[key] for key in ['input_collections', 'id', 'current_state']}
            if index == 'output_state':
                startkey['output_collections'] = dbitem['output_collections']
            resp = self.query(collection, state, since=since, index=index, Limit=limit, ExclusiveStartKey=startkey)
        else:
            resp = self.query(collection, state, since=since, index=index, Limit=limit)
        for i in resp['Items']:
            items['items'].append(self.dbitem_to_item(i))
        if 'LastEvaluatedKey' in resp:
            items['nextkey'] = self.key_to_catid(resp['LastEvaluatedKey'])
        return items

    def get_items(self, *args, limit=None, **kwargs) -> Dict:
        """Get items from database

        Args:
            limit (int, optional): Maximum number of items to return. Defaults to None.

        Returns:
            Dict: StateDB Items
        """
        resp = self.get_items_page(*args, **kwargs)
        items = resp['items']
        logger.debug(f"Fetched page of {len(items)} items from statedb")
        while 'nextkey' in resp and (limit is None or len(items) < limit):
            resp = self.get_items_page(*args, nextkey=resp['nextkey'], **kwargs)
            logger.debug(f"Fetched page of {len(resp['items'])} items from statedb")
            items += resp['items']
        if limit is None or len(items) < limit:
            return items
        return items[:limit]

    def get_state(self, catid: str) -> str:
        """Get current state of Item

        Args:
            catid (str): The catalog ID

        Returns:
            str: Current state: QUEUED, PROCESSING, COMPLETED, FAILED, INVALID
        """
        response = self.table.get_item(Key=self.catid_to_key(catid))
        if 'Item' in response:
            return response['Item']['current_state'].split('_')[0]
        else:
            # assuming no such item in database
            return ""

    def get_states(self, catids: List[str]) -> Dict[str, str]:
        """Get current state of items

        Args:
            catids (List[str]): List of catalog IDs

        Returns:
            Dict[str, str]: Dictionary of catalog IDs to state
        """
        states = {}
        for dbitem in self.get_dbitems(catids):
            item = self.dbitem_to_item(dbitem)
            states[item['catid']] = item['state']
        return states

    def set_processing(self, catid: str, execution: str) -> str:
        """Set Item to PROCESSING

        Args:
            catid (str): A Cirrus catalog
            execution (str): An ARN to the State Machine execution

        Returns:
            str: DynamoDB response
        """
        response = self.table.update_item(
            Key=self.catid_to_key(catid),
            UpdateExpression='SET current_state=:p, execution=:exe',
            ExpressionAttributeValues={
                ':p': f"PROCESSING_{datetime.now().isoformat()}",
                ':exe': execution
            }
        )
        return response

    def set_completed(self, catid: str, urls: List[str]) -> str:
        """Set this catalog as COMPLETED

        Args:
            catid (str): The Cirrus Catalog
            urls (List[str]): A list of output URLs to STAC Items

        Returns:
            str: DynamoDB response
        """
        response = self.table.update_item(
            Key=self.catid_to_key(catid),
            UpdateExpression='SET current_state=:p, output_urls=:urls',
            ExpressionAttributeValues={
                ':p': f"COMPLETED_{datetime.now().isoformat()}",
                ':urls': urls
            }
        )
        return response

    def set_failed(self, catid: str, msg: str) -> str:
        """Set this catalog as FAILED

        Args:
            catid (str): The Cirrus Catalog
            msg (str): An error message to include in DynamoDB Item

        Returns:
            str: DynamoDB response
        """
        response = self.table.update_item(
            Key=self.catid_to_key(catid),
            UpdateExpression='SET current_state=:p, error_message=:err',
            ExpressionAttributeValues={
                ':p': f"FAILED_{datetime.now().isoformat()}",
                ':err': msg
            }
        )
        return response

    def set_invalid(self, catid: str, msg: str) -> str:
        """Set this catalog as INVALID

        Args:
            catid (str): The Cirrus Catalog
            msg (str): An error message to include in DynamoDB Item

        Returns:
            str: DynamoDB response
        """
        response = self.table.update_item(
            Key=self.catid_to_key(catid),
            UpdateExpression='SET current_state=:p, error_message=:err',
            ExpressionAttributeValues={
                ':p': f"INVALID_{datetime.now().isoformat()}",
                ':err': msg
            }
        )
        return response

    def query(self, collection: str, state: str=None, since: str=None,
                     index: str='input_state', select: str='ALL_ATTRIBUTES', **kwargs) -> Dict:
        """Perform a single Query on a DynamoDB index

        Args:
            collection (str): /-separated ordered list of collections
            state (str, optional): The state to query. Defaults to None.
            since (str, optional): Query for items since this time. Defaults to None.
            index (str, optional): The DynamoDB index to query (input_state, output_state). Defaults to 'input_state'.
            select (str, optional): DynamoDB Select statement (ALL_ATTRIBUTES, COUNT). Defaults to 'ALL_ATTRIBUTES'.

        Returns:
            Dict: DynamoDB response
        """
        expr = Key(INDEX_KEYS[index]).eq(collection)
        if state and since:
            start = datetime.now() - self.since_to_timedelta(since)
            begin = f"{state}_{start.isoformat()}"
            end = f"{state}_{datetime.now().isoformat()}"
            expr = expr & Key('current_state').between(begin, end)
        elif state:
            expr = expr & Key('current_state').begins_with(state)
        resp = self.table.query(IndexName=index, KeyConditionExpression=expr, Select=select, **kwargs)
        return resp

    @classmethod
    def catid_to_key(cls, catid: str) -> Dict:
        """Create DynamoDB Key from catalog ID

        Args:
            catid (str): The catalog ID

        Returns:
            Dict: Dictionary containing the DynamoDB Key
        """
        parts = catid.split('/workflow-')
        key = {
            'input_collections': parts[0],
            'id': parts[1]
        }
        return key

    @classmethod
    def key_to_catid(cls, key: Dict) -> str:
        """Get catalog ID given a DynamoDB Key

        Args:
            key (Dict): DynamoDB Key

        Returns:
            str: Catalog ID
        """
        return f"{key['input_collections']}/workflow-{key['id']}"

    @classmethod
    def get_input_catalog_url(self, dbitem):
        catid = self.key_to_catid(dbitem)
        return f"s3://{CATALOG_BUCKET}/{catid}/input.json"

    @classmethod
    def dbitem_to_item(cls, dbitem: Dict, region: str=os.getenv('AWS_REGION', 'us-west-2')) -> Dict:
        state, updated_at = dbitem['current_state'].split('_')
        workflow = dbitem['id'].split('/')[0]
        item = {
            "catid": cls.key_to_catid(dbitem),
            "workflow": workflow,
            "input_collections": dbitem['input_collections'],
            "output_collections": dbitem['output_collections'],
            "state": state,
            "created_at": dbitem['created_at'],
            "updated_at": updated_at,
            "input_catalog": cls.get_input_catalog_url(dbitem)
        }
        if 'execution' in dbitem:
            exe_url = f"https://{region}.console.aws.amazon.com/states/home?region={region}#/executions/details/{dbitem['execution']}"
            item['execution'] = exe_url
        if 'error_message' in dbitem:
            item['error'] = dbitem['error_message']
        if 'output_urls' in dbitem:
            item['items'] = dbitem['output_urls']
        return item

    @classmethod
    def since_to_timedelta(cls, since: str) -> timedelta:
        """Convert a `since` field to a timedelta.

        Args:
            since (str): Contains an integer followed by a unit letter: 'd' for days, 'h' for hours, 'm' for minutes

        Returns:
            timedelta: [description]
        """
        unit = since[-1]
        # days, hours, or minutes
        assert(unit in ['d', 'h', 'm'])
        days = int(since[0:-1]) if unit == 'd' else 0
        hours = int(since[0:-1]) if unit == 'h' else 0
        minutes = int(since[0:-1]) if unit == 'm' else 0
        return timedelta(days=days, hours=hours, minutes=minutes)



# import moto before any boto3 module
from moto import mock_dynamodb2
import boto3
import inspect
import json
import os
import pytest
import unittest

from copy import deepcopy
from datetime import datetime
from decimal import Decimal
from cirruslib.statedb import StateDB, STATES


## fixtures
testpath = os.path.dirname(__file__)
table_name = 'cirrus-test-state'

test_item = {
    "id": "input-collection/workflow-test/item",
    "process": {
        "output_options": {
            "collections": {
                "output-collection": ".*"
            }
        }
    }
}

test_dbitem = {
    'input_collections': 'input-collection',
    'output_collections': 'output-collection',
    'id': 'test/item',
    'current_state': f"QUEUED_{datetime.now()}",
    'created_at': datetime.now()
}


def setUpModule():
    client = boto3.resource('dynamodb')
    with open(os.path.join(testpath, 'statedb_schema.json')) as f:
        schema = json.loads(f.read())
    table = client.create_table(**schema)
    table.meta.client.get_waiter('table_exists').wait(TableName=table_name)


def tearDownModule():
    client = boto3.resource('dynamodb')
    table = client.Table(table_name)
    table.delete()
    table.wait_until_not_exists()


class TestClassMethods(unittest.TestCase):

    testkey = {
        'input_collections': 'input-collection',
        'id': 'test/item'
    }

    def test_catid_to_key(self):
        key = StateDB.catid_to_key(test_item['id'])
        assert(key['input_collections'] == "input-collection")
        assert(key['id'] == 'test/item')

    def test_key_to_catid(self):
        catid = StateDB.key_to_catid(self.testkey)
        assert(catid == test_item['id'])

    def test_get_input_catalog_url(self):
        url = StateDB.get_input_catalog_url(self.testkey)
        assert(f"{test_item['id']}/input.json" in url)

    def test_dbitem_to_item(self):
        item = StateDB.dbitem_to_item(test_dbitem)
        assert(item['catid'] == test_item['id'])
        assert(item['workflow'] == 'test')
        assert(item['state'] == 'QUEUED')

    def test_since_to_timedelta(self):
        td = StateDB.since_to_timedelta('1d')
        assert(td.days == 1)
        td = StateDB.since_to_timedelta('1h')
        assert(td.seconds == 3600)
        td = StateDB.since_to_timedelta('10m')
        assert(td.seconds == 600)


# TODO - figure out why mocking still sends queries to AWS
#@mock_dynamodb2
class TestDbItems(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.statedb = StateDB(table_name)
        for i in range(10):
            newitem = deepcopy(test_item)
            newitem['id'] = newitem['id'] + str(i)
            cls.statedb.create_item(newitem)

    @classmethod
    def tearDownClass(cls):
        for i in range(10):
            cls.statedb.delete_item(test_item['id'] + str(i))

    def test_create_item(self):
        resp = self.statedb.create_item(test_item)
        assert(resp['ResponseMetadata']['HTTPStatusCode'] == 200)
        dbitem = self.statedb.get_dbitem(test_item['id'])
        assert(dbitem['id'] == test_dbitem['id'])
        self.statedb.delete_item(test_item['id'])
        #with self.assertRaises(Exception):
        #    dbitem = self.statedb.get_dbitem(test_item['id'])

    def test_get_dbitem(self):
        item = self.statedb.get_dbitem(test_item['id'] + '0')
        assert(item['id'] == test_dbitem['id'] + '0')
        assert(item['input_collections'] == test_dbitem['input_collections'])
        assert(item['current_state'].startswith('QUEUED'))

    def test_get_dbitems(self):
        ids = [test_item['id'] + str(i) for i in range(10)]
        dbitems = self.statedb.get_dbitems(ids)
        assert(len(dbitems) == len(ids))
        for dbitem in dbitems:
            assert(self.statedb.key_to_catid(dbitem) in ids)


class TestStates(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.statedb = StateDB(table_name)
        cls.statedb.create_item(test_item, state='QUEUED')
        for i, s in enumerate(STATES):
            newitem = deepcopy(test_item)
            newitem['id'] = newitem['id'] + str(i)
            cls.statedb.create_item(newitem, state=s)

    @classmethod
    def tearDownClass(cls):
        cls.statedb.delete_item(test_item['id'])
        for i in range(len(STATES)):
            cls.statedb.delete_item(test_item['id'] + str(i))        

    def test_get_state(self):
        for i, s in enumerate(STATES):
            state = self.statedb.get_state(test_item['id'] + str(i))
            assert(state == s)

    def test_get_states(self):
        ids = [test_item['id'] + str(i) for i in range(len(STATES))]
        states = self.statedb.get_states(ids)
        assert(len(ids) == len(states))
        for i, id in enumerate(ids):
            assert(states[id] == STATES[i])

    def test_set_processing(self):
        resp = self.statedb.set_processing(test_item['id'], execution='testarn')
        assert(resp['ResponseMetadata']['HTTPStatusCode'] == 200)
        dbitem = self.statedb.get_dbitem(test_item['id'])
        assert(dbitem['current_state'].startswith('PROCESSING'))
        assert(dbitem['execution'] == 'testarn')

    def test_set_complete(self):
        resp = self.statedb.set_completed(test_item['id'], urls=['output-url'])
        assert(resp['ResponseMetadata']['HTTPStatusCode'] == 200)
        dbitem = self.statedb.get_dbitem(test_item['id'])
        assert(dbitem['current_state'].startswith('COMPLETED'))
        assert(dbitem['output_urls'][0] == 'output-url')

    def test_set_failed(self):
        resp = self.statedb.set_failed(test_item['id'], msg='test failure')
        assert(resp['ResponseMetadata']['HTTPStatusCode'] == 200)
        dbitem = self.statedb.get_dbitem(test_item['id'])
        assert(dbitem['current_state'].startswith('FAILED'))

    def test_set_invalid(self):
        resp = self.statedb.set_invalid(test_item['id'], msg='test failure')
        assert(resp['ResponseMetadata']['HTTPStatusCode'] == 200)
        dbitem = self.statedb.get_dbitem(test_item['id'])
        assert(dbitem['current_state'].startswith('INVALID'))

# TODO - figure out why mocking still sends queries to AWS
#@mock_dynamodb2
class TestItems(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.statedb = StateDB(table_name)
        cls.statedb.create_item(test_item, state='QUEUED')
        for i, s in enumerate(STATES):
            newitem = deepcopy(test_item)
            newitem['id'] = newitem['id'] + str(i)
            cls.statedb.create_item(newitem, state=s)

    @classmethod
    def tearDownClass(cls):
        cls.statedb.delete_item(test_item['id'])
        for i in range(len(STATES)):
            cls.statedb.delete_item(test_item['id'] + str(i))     


from __future__ import annotations

import boto3
import json
import logging
import os
import re
import uuid

from boto3utils import s3
from cirruslib.statedb import StateDB
from cirruslib.transfer import get_s3_session
from cirruslib.utils import get_path
from typing import Dict, Optional, List

# envvars
LOG_LEVEL = os.getenv('CIRRUS_LOG_LEVEL', 'INFO')
DATA_BUCKET = os.getenv('CIRRUS_DATA_BUCKET', None)
CATALOG_BUCKET = os.getenv('CIRRUS_CATALOG_BUCKET', None)
PROCESS_QUEUE = os.getenv('CIRRUS_PROCESS_QUEUE', None)
PUBLISH_TOPIC_ARN = os.getenv('CIRRUS_PUBLISH_TOPIC_ARN', None)

logger = logging.getLogger(__name__)
logger.setLevel(LOG_LEVEL)

# clients
statedb = StateDB()
sqsclient = boto3.client('sqs')
snsclient = boto3.client('sns')
sqs_url = None
if PROCESS_QUEUE:
    sqs_url = sqsclient.get_queue_url(QueueName=PROCESS_QUEUE)['QueueUrl']


class Catalog(dict):

    def __init__(self, *args, state_item=None, **kwargs):
        """Initialize a Catalog, verify required fields, and assign an ID

        Args:
            state_item (Dict, optional): Dictionary of entry in StateDB. Defaults to None.
        """
        super(Catalog, self).__init__(*args, **kwargs)

        # validate process block
        assert(self['type'] == 'FeatureCollection')
        assert('process' in self)
        assert('output_options' in self['process'])
        assert('collections' in self)
        assert('workflow' in self['process'])

        # convert old functions field to tasks
        if 'functions' in self['process']:
            self['process']['tasks'] = self['process'].pop('functions')

        assert('tasks' in self['process'])

        # TODO - validate with a JSON schema
        #if schema:
        #    pass
        # For now, just make check that there is at least one item
        assert(len(self['features']) > 0)
        for item in self['features']:
            if 'links' not in item:
                item['links'] = []

        # Input collections
        if 'input_collections' not in self['process']:
            cols = sorted(list(set([i['collection'] for i in self['features'] if 'collection' in i])))
            self['process']['input_collections'] = cols if len(cols) != 0 else 'none'

        # generate ID
        collections_str = '/'.join(self['process']['input_collections'])
        items_str = '/'.join(sorted(list([i['id'] for i in self['features']])))
        self['id'] = f"{collections_str}/workflow-{self['process']['workflow']}/{items_str}"

        self.state_item = state_item

    # assign collections to Items given a mapping of Col ID: ID regex
    def assign_collections(self):
        """Assign new collections to all Items (features) in Catalog
            based on self['process']['output_options']['collections']
        """
        collections = self['process']['output_options']['collections']
        for item in self['features']:
            for col in collections:
                regex = re.compile(collections[col])
                if regex.match(item['id']):
                    logger.debug(f"Setting {item['id']} collection to {col}")
                    item['collection'] = col

    def get_payload(self) -> Dict:
        """Get original payload for this Catalog

        Returns:
            Dict: Cirrus Input Catalog
        """
        payload = json.dumps(self)
        if CATALOG_BUCKET and len(payload.encode('utf-8')) > 30000:
            assert(CATALOG_BUCKET)
            url = f"s3://{CATALOG_BUCKET}/payloads/{uuid.uuid1()}.json"
            s3().upload_json(self, url)
            return {'url': url}
        else:
            return dict(self)

    # publish the items in this catalog
    def publish_to_s3(self, bucket, public=False) -> List:
        """Publish all Items to s3

        Args:
            bucket (str): Name of bucket to publish to
            public (bool, optional): Make published STAC Item public. Defaults to False.

        Returns:
            List: List of s3 URLs to published Items
        """
        opts = self['process'].get('output_options', {})
        s3urls = []
        for item in self['features']:
            # determine URL of data bucket to publish to- always do this
            url = os.path.join(get_path(item, opts.get('path_template')), f"{item['id']}.json")
            if url[0:5] != 's3://':
                url = f"s3://{bucket}/{url.lstrip('/')}"

            # add self link (and remove existing self link if present)
            item['links'] = [l for l in item['links'] if l['rel'] != 'self']
            item['links'].insert(0, {
                'rel': 'self',
                'href': url,
                'type': 'application/json'
            })

            # get s3 session
            s3session = get_s3_session(s3url=url)

            # publish to bucket
            headers = opts.get('headers', {})
            
            extra = {'ContentType': 'application/json'}
            extra.update(headers)
            s3session.upload_json(item, url, public=public, extra=extra)
            s3urls.append(url)
            logger.info(f"Uploaded STAC Item {item['id']} as {url}")

        return s3urls

    @classmethod
    def sns_attributes(self, item) -> Dict:
        """Create attributes from Item for publishing to SNS

        Args:
            item (Dict): A STAC Item

        Returns:
            Dict: Attributes for SNS publishing
        """
        attr = {
            'collection': {
                'DataType': 'String',
                'StringValue': item['collection']
            },
            'datetime': {
                'DataType': 'String',
                'StringValue': item['properties']['datetime']
            },
            'bbox.ll_lon': {
                'DataType': 'Number',
                'StringValue': str(item['bbox'][0])
            },
            'bbox.ll_lat': {
                'DataType': 'Number',
                'StringValue': str(item['bbox'][1])
            },
            'bbox.ur_lon': {
                'DataType': 'Number',
                'StringValue': str(item['bbox'][2])
            },
            'bbox.ur_lat': {
                'DataType': 'Number',
                'StringValue': str(item['bbox'][3])
            }     
        }
        if 'eo:cloud_cover' in item['properties']:
            attr['cloud_cover'] = {
                'DataType': 'Number',
                'StringValue': str(item['properties']['eo:cloud_cover'])
            }
        return attr

    def publish_to_sns(self, topic_arn=PUBLISH_TOPIC_ARN):
        """Publish this catalog to SNS

        Args:
            topic_arn (str, optional): ARN of SNS Topic. Defaults to PUBLISH_TOPIC_ARN.
        """
        for item in self['features']:
            logger.info(f"Publishing item {item['id']} to {topic_arn}")
            response = snsclient.publish(TopicArn=topic_arn, Message=json.dumps(item),
                                        MessageAttributes=self.sns_attributes(item))
            logger.debug(f"Response: {json.dumps(response)}")           

    def process(self) -> str:
        """Add this catalog to procesing queue

        Returns:
            str: Catalog ID
        """
        assert(CATALOG_BUCKET)
        assert(sqs_url)

        # create DynamoDB record - this will always overwrite any existing process
        try:
            statedb.create_item(self) 
        except Exception as err:
            msg = f"Error adding {self['id']} to database ({err})"
            logger.error(msg)
            return

        try:
            url = f"s3://{CATALOG_BUCKET}/{self['id']}/input.json"
            s3().upload_json(self, url)
            logger.debug(f"Uploaded {url}")
        except Exception as err:
            msg = f"Error adding {self['id']} input catalog to s3 ({err})"
            logger.error(msg)
            return

        try:
            response = sqsclient.send_message(
                QueueUrl = sqs_url,
                MessageBody = json.dumps(self)
            )
            # TODO - check response
            logger.info(f"Queued {self['id']}")
            return self['id']
        except Exception as err:
            msg = f"queue: error queuing {self['id']} ({err})"
            logger.error(msg)
            statedb.set_failed(self['id'], msg)    

        return self['id'] 


class Catalogs(object):

    def __init__(self, catalogs, state_items=None):
        self.catalogs = catalogs
        if state_items:
            assert(len(state_items) == len(self.catalogs))
        self.state_items = state_items

    def __getitem__(self, index):
        return self.catalogs[index]

    @property
    def catids(self) -> List[str]:
        """Return list of catalog IDs

        Returns:
            List[str]: List of Catalog IDs
        """
        return [c['id'] for c in self.catalogs]

    @classmethod
    def from_payload(cls, payload: Dict, **kwargs) -> Catalogs:
        """Parse a Cirrus payload and return a Catalogs instance

        Args:
            payload (Dict): A payload from SNS, SQS, or containing an s3 URL to payload

        Returns:
            Catalogs: A Catalogs instance
        """
        catalogs = []
        if 'Records' in payload:
            for record in [json.loads(r['body']) for r in payload['Records']]:
                if 'Message' in record:
                    # SNS
                    cat = Catalog(json.loads(record['Message']))
                    catalogs.append(cat)
                else:
                    # SQS
                    catalogs.append(Catalog(record))
        elif 'url' in payload:
            catalogs = [Catalog(s3().read_json(payload['url']))]
        elif 'Parameters' in payload and 'url' in payload['Parameters']:
            catalogs = [Catalog(s3().read_json(payload['Parameters']['url']))]
        else:
            catalogs = [Catalog(payload)]
        return cls(catalogs)

    @classmethod
    def from_catids(cls, catids: List[str], **kwargs) -> Catalogs:
        """Create Catalogs from list of Catalog IDs

        Args:
            catids (List[str]): List of catalog IDs

        Returns:
            Catalogs: A Catalogs instance
        """
        items = [statedb.dbitem_to_item(statedb.get_dbitem(catid)) for catid in catids]
        catalogs = []
        for item in items:
            cat = Catalog(s3().read_json(item['input_catalog']))
            catalogs.append(cat)
        logger.debug(f"Retrieved {len(catalogs)} from state db")
        return cls(catalogs, state_items=items)

    """
    @classmethod
    def from_statedb_paged(cls, collections, state, since: str=None, index: str='input_state', limit=None):
        catalogs = []
        # get first page
        resp = statedb.get_items_page(collections, state, since, index)
        for it in resp['items']:
            cat = Catalog(s3().read_json(it['input_catalog']))
            catalogs.append(cat)
        logger.debug(f"Retrieved {len(catalogs)} from state db")
        yield cls(catalogs, state_items=resp['items'])
        catalogs = []
        while 'nextkey' in resp:
            resp = statedb.get_items_page(collections, state, since, index, nextkey=resp['nextkey'])
            for it in resp['items']:
                cat = Catalog(s3().read_json(it['input_catalog']))
                catalogs.append(cat)
            logger.debug(f"Retrieved {len(catalogs)} from state db")
            yield cls(catalogs, state_items=resp['items'])
    """

    @classmethod
    def from_statedb(cls, collections, state, since: str=None, index: str='input_state', limit=None) -> Catalogs:
        """Create Catalogs object from set of StateDB Items

        Args:
            collections (str): String of collections (input or output depending on `index`)
            state (str): The state (QUEUED, PROCESSING, COMPLETED, FAILED, INVALID) of StateDB Items to get
            since (str, optional): Get Items since this duration ago (e.g., 10m, 8h, 1w). Defaults to None.
            index (str, optional): 'input_state' or 'output_state' Defaults to 'input_state'.
            limit ([type], optional): Max number of Items to return. Defaults to None.

        Returns:
            Catalogs: Catalogs instance
        """
        catalogs = []
        items = statedb.get_items(collections, state, since, index, limit=limit)
        logger.debug(f"Retrieved {len(items)} total items from statedb")
        for item in items:
            cat = Catalog(s3().read_json(item['input_catalog']))
            catalogs.append(cat)
        logger.debug(f"Retrieved {len(catalogs)} input catalogs")
        return cls(catalogs, state_items=items)

    def get_states(self):
        if self.state_items is None:
            items = [statedb.dbitem_to_item(i) for i in statedb.get_dbitems(self.catids)]
            self.state_items = items
        states = {c['catid']: c['state'] for c in self.state_items}
        return states

    def process(self, replace=False):
        """Create Item in Cirrus State DB for each Catalog and add to processing queue

        Args:
            catalog (Dict): A Cirrus Input Catalog
        """
        # check existing states            
        states = self.get_states()
        catalogs = []
        for cat in self.catalogs:
            # check existing state for Item, if any
            state = states.get(cat['id'], '')
            # don't try and process these - if they are stuck they should be removed from db
            #if state in ['QUEUED', 'PROCESSING']:
            #    logger.info(f"Skipping {cat['id']}, in {state} state")
            #    continue
            _replace = replace or cat['process'].get('replace', False)
            if state in ['FAILED', ''] or _replace:
                catalogs.append(cat)
            else:
                logger.info(f"Skipping {cat['id']}, in {state} state")
                continue

        # add to database, to s3, then to queue
        catids = []
        for catalog in catalogs:
            catids.append(catalog.process())
        return catids


import boto3
import json
import logging
import os

from boto3utils import s3
from typing import Dict, Optional, List

from pystac import STAC_IO, Catalog, CatalogType, Collection

logger = logging.getLogger(__name__)
logger.setLevel(os.getenv('CIRRUS_LOG_LEVEL', 'INFO'))


# envvars
DATA_BUCKET = os.getenv('CIRRUS_DATA_BUCKET', None)
PUBLIC_CATALOG = os.getenv('CIRRUS_PUBLIC_CATALOG', False)
STAC_VERSION = os.getenv('CIRRUS_STAC_VERSION', '1.0.0-beta.2')
DESCRIPTION = os.getenv('CIRRUS_STAC_DESCRIPTION', 'Cirrus STAC')
AWS_REGION = os.getenv('AWS_REGION')

ROOT_URL = f"s3://{DATA_BUCKET}/catalog.json"


def s3stac_read(uri):
    if uri.startswith('s3'):
        return json.dumps(s3().read_json(uri))
    else:
        return STAC_IO.default_read_text_method(uri)

def s3stac_write(uri, txt):
    extra = {
        'ContentType': 'application/json'
    }
    if uri.startswith('s3'):
        s3().upload_json(json.loads(txt), uri, extra=extra, public=PUBLIC_CATALOG)
    else:
        STAC_IO.default_write_text_method(uri, txt)

STAC_IO.read_text_method = s3stac_read
STAC_IO.write_text_method = s3stac_write


def get_root_catalog() -> Dict:
    """Get Cirrus root catalog from s3

    Returns:
        Dict: STAC root catalog
    """
    if s3().exists(ROOT_URL):
        cat = Catalog.from_file(ROOT_URL)
    else:
        catid = DATA_BUCKET.split('-data-')[0]
        cat = Catalog(id=catid, description=DESCRIPTION)
    logger.debug(f"Fetched {cat.describe()}")
    return cat


# add this collection to Cirrus catalog
def add_collection(collection):
    cat = get_root_catalog()
    col = Collection(**collection)
    cat.add_child(col)
    cat.normalize_and_save(ROOT_URL, CatalogType=CatalogType.ABSOLUTE_PUBLISHED)
    return cat


# import moto before any boto3 module
from moto import mock_s3
import boto3
import os
import unittest

from boto3utils import s3
from cirruslib import transfer
from shutil import rmtree

testpath = f"{os.path.dirname(__file__)}/test_transfer"
testbucket = 'testbucket'

@mock_s3
class Test(unittest.TestCase):

    def setUp(self):
        client = s3()
        client.s3.create_bucket(Bucket=testbucket)
        client.s3.put_object(Body='test', Bucket=testbucket, Key=os.path.basename(__file__))
        os.makedirs(testpath, exist_ok=True)
        #client.upload_file(Filename=os.path.join(testpath, 'test.json'), Bucket=testbucket, Key='test.json')
        #yield client

    @classmethod
    def tearDownClass(cls):
        rmtree(testpath)

    def get_test_item(self):
        item = {
            'id': 'test-item',
            'collection': 'test-collection',
            'assets': {
                'local': {
                    'href': __file__
                },
                'remote': {
                    'href': f"s3://{testbucket}/{os.path.basename(__file__)}"
                }
            }
        }
        return item

    def test_get_s3_session(self):
        session = transfer.get_s3_session(region_name='us-west-2')
        buckets = session.s3.list_buckets()
        assert(buckets['Buckets'][0]['Name'] == testbucket)

    def test_download_from_http(self):
        url = 'https://raw.githubusercontent.com/matthewhanson/boto3-utils/master/README.md'
        fname = transfer.download_from_http(url, path=testpath)
        assert(os.path.exists(fname))
        with open(fname) as f:
            lines = f.readlines()
        assert('boto3-utils') in lines[-1]

    def test_download_item_assets(self):
        item = self.get_test_item()
        new_item = transfer.download_item_assets(item, path=testpath)
        for k in new_item['assets']:
            assert(os.path.exists(new_item['assets'][k]['href']))

    def test_upload_item_assets(self):
        item = self.get_test_item()
        path_template = 's3://testbucket/${id}/test'
        assets = ['local']
        new_item = transfer.upload_item_assets(item, assets=assets, path_template=path_template, s3_urls=True, region_name='us-west-2')
        for k in assets:
            assert(new_item['assets'][k]['href'].startswith('s3://'))
            assert(s3().exists(new_item['assets'][k]['href']))

    def test_get_path(self):
        path_template = 's3://testbucket/${collection}/${id}'
        item = self.get_test_item()
        path = transfer.get_path(self.get_test_item(), template = path_template)
        assert(path == f"s3://{testbucket}/{item['collection']}/{item['id']}")


import boto3
import json
import logging
import requests
import uuid

from boto3utils import s3
from dateutil.parser import parse as dateparse
from os import getenv
from string import Formatter, Template
from typing import Dict, Optional, List

# configure logger - CRITICAL, ERROR, WARNING, INFO, DEBUG
logger = logging.getLogger(__name__)
logger.setLevel(getenv('CIRRUS_LOG_LEVEL', 'INFO'))

batch_client = boto3.client('batch')


def submit_batch_job(payload, arn, queue='basic-ondemand', definition='geolambda-as-batch'):
    # envvars
    STACK_PREFIX = getenv('CIRRUS_STACK')
    CATALOG_BUCKET = getenv('CIRRUS_CATALOG_BUCKET')

    # upload payload to s3
    url = f"s3://{CATALOG_BUCKET}/batch/{uuid.uuid1()}.json"
    s3().upload_json(payload, url)
    kwargs = {
        'jobName': arn.split(':')[-1],
        'jobQueue': f"{STACK_PREFIX}-{queue}",
        'jobDefinition': f"{STACK_PREFIX}-{definition}",
        'parameters': {
            'lambda_function': arn,
            'url': url
        },
        'containerOverrides': {
            'vcpus': 1,
            'memory': 512,
        }
    }
    logger.debug(f"Submitted batch job with payload {url}")
    response = batch_client.submit_job(**kwargs)
    logger.debug(f"Batch response: {response}")


def get_path(item: Dict, template: str='${collection}/${id}') -> str:
    """Get path name based on STAC Item and template string

    Args:
        item (Dict): A STAC Item.
        template (str, optional): Path template using variables referencing Item fields. Defaults to '${collection}/${id}'.

    Returns:
        [str]: A path name
    """
    _template = template.replace(':', '__colon__')
    subs = {}
    for key in [i[1] for i in Formatter().parse(_template.rstrip('/')) if i[1] is not None]:
        # collection
        if key == 'collection':
            subs[key] = item['collection']
        # ID
        elif key == 'id':
            subs[key] = item['id']
        # derived from date
        elif key in ['year', 'month', 'day']:
            dt = dateparse(item['properties']['datetime'])
            vals = {'year': dt.year, 'month': dt.month, 'day': dt.day}
            subs[key] = vals[key]
        # Item property
        else:
            subs[key] = item['properties'][key.replace('__colon__', ':')]
    return Template(_template).substitute(**subs).replace('__colon__', ':')